var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
import { NotFoundError } from '@nodescript/errors';
import { Logger } from '@nodescript/logger';
import { createServer } from 'http';
import { config } from 'mesh-config';
import { dep } from 'mesh-ioc';
import { HttpContext } from './HttpContext.js';
export class HttpServer {
    constructor() {
        this.server = null;
        this.requestsPerSocket = new Map();
        this.stopping = false;
        this.config = {
            port: this.HTTP_PORT,
            address: this.HTTP_ADDRESS,
            socketTimeout: this.HTTP_TIMEOUT,
            shutdownDelay: this.HTTP_SHUTDOWN_DELAY,
            requestBodyLimitBytes: this.HTTP_REQUEST_BODY_LIMIT_BYTES,
        };
    }
    async start() {
        if (this.server) {
            return;
        }
        this.stopping = false;
        this.requestsPerSocket.clear();
        const server = createServer((req, res) => this.handleRequest(req, res));
        this.server = server;
        server.on('connection', sock => this.onConnection(sock));
        server.on('request', (req, res) => this.onRequest(req, res));
        server.setTimeout(this.config.socketTimeout);
        await new Promise((resolve, reject) => {
            server.on('error', err => reject(err));
            server.listen(this.config.port, this.config.address || undefined, () => {
                this.logger.info(`${this.constructor.name}: listening on ${this.config.port}`);
                resolve();
            });
        });
    }
    async stop() {
        const server = this.server;
        if (!server) {
            return;
        }
        this.stopping = true;
        // This is required in environments like K8s where traffic is still being sent after SIGTERM
        const waitSec = Math.round(this.config.shutdownDelay / 1000);
        this.logger.info(`${this.constructor.name}: stop initiated, waiting for ${waitSec}s before stopping to accept new connections`);
        await new Promise(r => setTimeout(r, this.config.shutdownDelay));
        this.logger.info(`${this.constructor.name}: waiting for existing requests to finish`);
        const closePromise = new Promise((resolve, reject) => server.close(err => err ? reject(err) : resolve()));
        const timeout = setTimeout(() => this.destroyAllSockets(), this.config.socketTimeout);
        this.closeIdleSockets();
        await closePromise;
        clearTimeout(timeout);
        this.server = null;
    }
    getServer() {
        return this.server;
    }
    async handleRequest(req, res) {
        try {
            const ctx = new HttpContext(this.config, req, res);
            await this.handle(ctx, () => {
                throw new NotFoundError(`${ctx.method} ${ctx.url.pathname}`);
            });
            ctx.sendResponse();
        }
        catch (error) {
            // Minimal error handling here, should be implemented by error handler
            const status = Number(error.status) || 500;
            this.logger.error(`${this.constructor.name}: request failed`, { error });
            res.writeHead(status, { 'content-type': 'application/json' });
            res.end(JSON.stringify({
                name: error.name,
                message: error.message,
            }));
        }
    }
    /**
     * Tracks the number of pending requests per socket, so that the server could close gracefully
     */
    onConnection(sock) {
        this.requestsPerSocket.set(sock, 0);
        sock.once('close', () => this.requestsPerSocket.delete(sock));
    }
    /**
     * Tracks requests per connection, close connection during shutdown, when requests are served.
     */
    onRequest(req, res) {
        const socket = req.socket ?? req.client;
        this.requestsPerSocket.set(socket, this.getSocketRequests(socket) + 1);
        res.once('finish', () => {
            const pending = this.getSocketRequests(socket) - 1;
            this.requestsPerSocket.set(socket, pending);
            if (this.stopping && pending === 0) {
                req.socket.end();
            }
        });
    }
    getSocketRequests(socket) {
        return this.requestsPerSocket.get(socket) ?? 0;
    }
    /**
     * Gracefully close idle sockets.
     */
    closeIdleSockets() {
        for (const [sock, count] of this.requestsPerSocket) {
            if (count === 0) {
                sock.end();
            }
        }
    }
    /**
     * Destroy all sockets non-gracefully.
     */
    destroyAllSockets() {
        for (const sock of this.requestsPerSocket.keys()) {
            sock.destroy();
        }
        this.requestsPerSocket.clear();
    }
}
__decorate([
    config({ default: 8080 }),
    __metadata("design:type", Number)
], HttpServer.prototype, "HTTP_PORT", void 0);
__decorate([
    config({ default: '' }),
    __metadata("design:type", String)
], HttpServer.prototype, "HTTP_ADDRESS", void 0);
__decorate([
    config({ default: 120_000 }),
    __metadata("design:type", Number)
], HttpServer.prototype, "HTTP_TIMEOUT", void 0);
__decorate([
    config({ default: 30_000 }),
    __metadata("design:type", Number)
], HttpServer.prototype, "HTTP_SHUTDOWN_DELAY", void 0);
__decorate([
    config({ default: 5 * 1024 * 1024 }),
    __metadata("design:type", Number)
], HttpServer.prototype, "HTTP_REQUEST_BODY_LIMIT_BYTES", void 0);
__decorate([
    dep(),
    __metadata("design:type", Logger)
], HttpServer.prototype, "logger", void 0);
//# sourceMappingURL=HttpServer.js.map